/*
 * Created 2023-07-05
 */
package com.nortal.vst.rc.demo;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;
import java.util.regex.Pattern;

import javax.net.ssl.SSLContext;

import org.apache.commons.io.IOUtils;
import org.apache.http.StatusLine;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.DefaultHostnameVerifier;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.conn.ssl.TrustSelfSignedStrategy;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.BasicHttpClientConnectionManager;
import org.apache.http.ssl.SSLContexts;

import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

/**
 * 
 */
@Log4j2
public class RestClientDemo implements AutoCloseable {

    private static final Pattern trailingSlashPattern = Pattern.compile("/+$");

    public static final String URL = "url";
    public static final String RICHTLINIE = "richtlinie";
    public static final String TRUSTSTORE_FILE = "truststore.file";
    public static final String TRUSTSTORE_PASSWORD = "truststore.password";
    public static final String KEYSTORE_FILE = "keystore.file";
    public static final String KEYSTORE_PASSWORD = "keystore.password";
    public static final String USE_V2_API = "use.v2.api";

    private static final int RESPONSE_STATUS_200 = 200;
    private static final int RESPONSE_STATUS_201 = 201;
    private static final int RESPONSE_STATUS_400 = 400;
    private static final int RESPONSE_STATUS_401 = 401;
    private static final int RESPONSE_STATUS_403 = 403;
    private static final int RESPONSE_STATUS_409 = 409;
    private static final int RESPONSE_STATUS_500 = 500;
    private static final int RESPONSE_STATUS_502 = 502;
    private static final int RESPONSE_STATUS_503 = 503;

    private CloseableHttpClient httpClient;
    private String requestUrl;
    private boolean useV2Api;

    public RestClientDemo(final Properties config) {
        httpClient = createHttpsClient(config);
        requestUrl = withoutTrailingSlash(config.getProperty(URL));
        useV2Api = Boolean.parseBoolean(config.getProperty(USE_V2_API));
        if (useV2Api) {
            requestUrl += "/v2";
        }
        requestUrl += "/" + config.getProperty(RICHTLINIE);
        log.info("posting delivery to URL {}", requestUrl);
    }

    @Override
    public void close() {
        if (httpClient != null) {
            IOUtils.closeQuietly(httpClient);
            httpClient = null;
        }
    }

    public static void main(final String[] args) throws IOException {
        String configFile = "src/main/resources/config.properties";

        log.info("Load configuration from file '{}'", configFile);
        Properties config = new Properties();
        try (FileInputStream fileIn = new FileInputStream(configFile)) {
            config.load(fileIn);
        }
        log.info("Configuration loaded");
        checkConfiguration(config);

        try (RestClientDemo restClientDemo = new RestClientDemo(config)) {
            Files.list(Paths.get("./files_to_send"))
                    .filter(file -> Files.isRegularFile(file)
                            && file.getFileName().toString().toLowerCase().endsWith(".xml"))
                    .forEach(file -> {
                        log.info("-----------------------------");
                        log.info("Send file '{}'", file.toString());
                        Path outputFile = Paths.get("./files_response").resolve(file.getFileName());
                        try (FileInputStream fileIn = new FileInputStream(file.toFile())) {
                            restClientDemo.post(fileIn, outputFile);
                        } catch (Exception e) {
                            log.error(e);
                        }
                    });
        }
    }

    private static void checkConfiguration(final Properties config) throws FileNotFoundException {
        checkPropertyIsDefined(config, URL);
        checkPropertyIsDefined(config, RICHTLINIE);
        checkPropertyIsDefined(config, KEYSTORE_FILE);
        checkPropertyIsDefined(config, KEYSTORE_PASSWORD);
        checkPropertyIsDefined(config, TRUSTSTORE_FILE);
        checkPropertyIsDefined(config, TRUSTSTORE_PASSWORD);
        checkFileExist(KEYSTORE_FILE, Paths.get(config.getProperty(KEYSTORE_FILE)));
        checkFileExist(TRUSTSTORE_FILE, Paths.get(config.getProperty(TRUSTSTORE_FILE)));
    }

    private static void checkPropertyIsDefined(final Properties config, final String propertyKey) {
        String value = config.getProperty(propertyKey);
        if (value == null || value.trim().length() == 0) {
            throw new RestClientDemoException("Property '" + propertyKey + "' was not configured");
        }
    }

    private static void checkFileExist(final String propertyKey, final Path file) throws FileNotFoundException {
        if (!Files.exists(file)) {
            throw new FileNotFoundException(
                    "File '" + file.toString() + "' for property '" + propertyKey + "' not found");
        }
    }

    public @NonNull void post(final InputStream inputStream, final Path outFile) {
        // can't use UriComponents without spring
        final HttpPost httpPost = new HttpPost(requestUrl);
        httpPost.addHeader("Content-Type", "text/plain");
        httpPost.addHeader("Accept", "text/plain");
        httpPost.setEntity(new InputStreamEntity(inputStream));
        try (FileOutputStream fileOut = new FileOutputStream(outFile.toFile());) {
            final CloseableHttpResponse res = httpClient.execute(httpPost);
            try (InputStream responseStream = res.getEntity().getContent()) {
                IOUtils.copy(responseStream, fileOut);
            }
            if (useV2Api) {
                handleResponseV2(res);
            } else {
                handleResponse(res);
            }
        } catch (Exception e) {
            throw new RestClientDemoException("request failed", e);
        }
    }

    private void handleResponse(final CloseableHttpResponse response) {
        final StatusLine statusLine = response.getStatusLine();
        final int statusCode = statusLine.getStatusCode();
        log.info("Response Status: {}; Content-Type: {}; Content-Length: {}; Content-Encoding: {}", statusCode,
                response.getEntity().getContentType(), response.getEntity().getContentLength(),
                response.getEntity().getContentEncoding());
        if (RESPONSE_STATUS_200 == statusCode) {
            // erfolgreich
            log.info("Daten erfolgreich uebermittelt.");

            // Daten wurden von der VST-PSN angenommen.
            // Die Daten waren in diesem Fall gueltig - eine manuelle
            // Auswertung des Response ist nicht zwingend notwendig.

        } else if (RESPONSE_STATUS_400 == statusCode) {
            // Client (Daten) Fehler
            log.error("Client Fehler - Daten wurden nicht uebermittelt.");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Die Daten waren in diesem Fall NICHT gueltig - eine manuelle
            // Auswertung des Responses IST zwingend notwendig um den Grund
            // fuer die Ablehnung festzustellen.

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_401 == statusCode) {
            // SSL Fehler
            log.error("SSL Fehler - Kein Client Zertifikat angegeben");

            // Keine Verbindung zur VST-PSN möglich.

        } else if (RESPONSE_STATUS_503 == statusCode) {
            // Server ueberlastet
            log.error("Server ueberlastet.");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // An dieser Stelle sollte der "Retry-After" Header ausgewertet
            // werden und die darin angegeben Sekunden abwarten vor dem
            // der Versand erneut versucht wird.

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_502 == statusCode) {
            // Keine freien Slots für die Verarbeitung
            log.error("Keine freien Slots für die Verarbeitung.");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Bitte liefern Sie die Datei zu einem späteren Zeitpunkt

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_500 == statusCode) {
            log.error("Schwerwiegender Fehler. " +
                    "Bitte dringend dem Vertrauensstelle Helpdesk (http://www.vertrauensstelle-gba.de/) melden.");

            // eigene Weiterverarbeitung hier einbauen.

        } else {
            // fehler aufgetretten
            log.error("Unbekanntes Ergebniss beim Restaufruf: {}", statusCode);
            log.error("Bitte dringend dem Vertrauensstelle Helpdesk (http://www.vertrauensstelle-gba.de/) melden.");

            // eigene Weiterverarbeitung hier einbauen.

        }
    }

    private void handleResponseV2(final CloseableHttpResponse response) {
        final StatusLine statusLine = response.getStatusLine();
        final int statusCode = statusLine.getStatusCode();
        log.info("Response Status: {}; Content-Type: {}; Content-Length: {}; Content-Encoding: {}", statusCode,
                response.getEntity().getContentType(), response.getEntity().getContentLength(),
                response.getEntity().getContentEncoding());
        if (RESPONSE_STATUS_200 == statusCode) {
            // erfolgreich
            log.info("Daten bereits bekannt.");

            // ACHTUNG: dies ist noch in der Umsetzung, zurzeit wird bei gleichen Daten ein 409 zurückgegeben
            // Daten sind in der VST schon bekannt, eine erneute Verarbeitung wurde nicht gestartet
        } else if (RESPONSE_STATUS_201 == statusCode) {
            // erfolgreich
            log.info("Daten erfolgreich uebermittelt.");

            // Daten wurden von der VST-PSN angenommen.
            // Die Daten waren in diesem Fall gueltig - eine manuelle
            // Auswertung des Response ist nicht zwingend notwendig.

        } else if (RESPONSE_STATUS_400 == statusCode) {
            // Client (Daten) Fehler
            log.error("Client Fehler - Daten waren fehlerhaft.");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Die Daten waren in diesem Fall NICHT gueltig - eine manuelle
            // Auswertung des Responses IST zwingend notwendig um den Grund
            // fuer die Ablehnung festzustellen.

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_401 == statusCode) {
            // SSL Fehler
            log.error("SSL Fehler - Kein Client Zertifikat angegeben");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Keine Verbindung zur VST-PSN möglich. Bitte Client Zertifikat prüfen.

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_403 == statusCode) {
            // Client (Daten) Fehler
            log.error("Client Fehler - nicht autorisierter Zugriff "
                    + "(falsches Zertifikat, nicht zugeordnet oder unbekannt)");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Bitte wenden Sie sich an unseren Support

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_409 == statusCode) {
            // Client (Daten) Fehler
            log.error("Client Fehler - Lieferung GUID bereits bekannt");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Bitte liefern Sie mit einer neuen GUID

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_503 == statusCode) {
            // Server ueberlastet
            log.error("Server ueberlastet.");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // An dieser Stelle sollte der "Retry-After" Header ausgewertet
            // werden und die darin angegeben Sekunden abwarten vor dem
            // der Versand erneut versucht wird.

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_502 == statusCode) {
            // Keine freien Slots für die Verarbeitung
            log.error("Keine freien Slots für die Verarbeitung.");

            // Daten wurden von der VST-PSN __NICHT__ angenommen.

            // Bitte liefern Sie die Datei zu einem späteren Zeitpunkt

            // eigene Weiterverarbeitung hier einbauen.

        } else if (RESPONSE_STATUS_500 == statusCode) {
            log.error("Schwerwiegender Fehler. " +
                    "Bitte dringend dem Vertrauensstelle Helpdesk (http://www.vertrauensstelle-gba.de/) melden.");

            // eigene Weiterverarbeitung hier einbauen.

        } else {
            // fehler aufgetretten
            log.error("Unbekanntes Ergebniss beim Restaufruf: {}", statusCode);
            log.error("Bitte dringend dem Vertrauensstelle Helpdesk (http://www.vertrauensstelle-gba.de/) melden.");

            // eigene Weiterverarbeitung hier einbauen.

        }
    }

    public static CloseableHttpClient createHttpsClient(final Properties config) {
        final HttpClientBuilder builder = HttpClients.custom();

        int timeout = 5000;
        RequestConfig requestConfig = RequestConfig.custom()
                // the time to establish the connection with the remote host
                .setConnectTimeout(timeout)
                // the time to wait for a connection from the connection manager/pool
                .setConnectionRequestTimeout(timeout)
                // the time waiting for data – after establishing the connection; maximum time of inactivity between two
                // data packets
                .setSocketTimeout(timeout * 12).build();
        builder.setDefaultRequestConfig(requestConfig);

        final File trustStoreFile = new File(config.getProperty(TRUSTSTORE_FILE));
        final File keyStoreFile = new File(config.getProperty(KEYSTORE_FILE));
        final char[] trustStorePassword = config.getProperty(TRUSTSTORE_PASSWORD).toCharArray();
        final char[] keyStorePassword = config.getProperty(KEYSTORE_PASSWORD).toCharArray();
        try {
            final SSLContext sslContext = SSLContexts.custom()
                    .loadTrustMaterial(trustStoreFile, trustStorePassword, new TrustSelfSignedStrategy())
                    .loadKeyMaterial(keyStoreFile, keyStorePassword, keyStorePassword)
                    .build();
            final SSLConnectionSocketFactory sslsf = new SSLConnectionSocketFactory(sslContext,
                    new DefaultHostnameVerifier());
            final Registry<ConnectionSocketFactory> socketFactoryRegistry = RegistryBuilder
                    .<ConnectionSocketFactory>create()
                    .register("https", sslsf)
                    .register("http", new PlainConnectionSocketFactory())
                    .build();
            final BasicHttpClientConnectionManager connectionManager = new BasicHttpClientConnectionManager(
                    socketFactoryRegistry);
            builder.setConnectionManager(connectionManager);
            return builder.build();
        } catch (Exception e) {
            throw new RestClientDemoException("could not create HTTPS client", e);
        }
    }

    public static String withoutTrailingSlash(final String string) {
        return trailingSlashPattern.matcher(string).replaceAll("");
    }
}
